# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

from libff.plugin import *


class Test(Plugin):
    """The "test" plugin offers some tests.
    """

    speed = Speed.SLOW

    use_cache = True

    attributes = [
        ("foo", String, "Some test data.")
    ]

    def can_handle(self, entry):
        return True

    def process(self, entry, cached):
        # pylint:disable=unspecified-encoding
        with open("/nonexistent"):
            pass
